//==============================================
//  check multiple stops for strings
//
//  Copyright (C)2019 Jörn Eichler (heuchi)
//  releases:
//  2022  D. Denolf (graffesmusic)- port to MuseScore 4
//  2024  FBXOPWKDOIR2 - added support for Cello and Viola
//  2024  D.Denolf (graffesmusic) - merged into 1 plugin
//  2024  D.Denolf (graffesmusic) - ported to MuseScore 4.4
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program.  If not, see <http://www.gnu.org/licenses/>.
//==============================================

// This is loosely based on the mechanism described
// by Ertuğrul Sevsay in his book on orchestration



import QtQuick 2.15
import QtQuick.Controls 2.15
import MuseScore 3.0

MuseScore {
	   	menuPath: "Plugins.Proof Reading.Check Violin Multiple Stops"
     	description: "Check if multiple stops are possible"
     	version: "0.2"
      	requiresScore: true;
      	
     	//4.4 title: qsTr("Check Multiple Stops")
        //4.4 thumbnailName: ""
        //4.4 categoryCode: "composing-arranging-tools"
        Component.onCompleted : {
			if (mscoreMajorVersion >= 4 && mscoreMinorVersion <= 3) {
				title = qsTr("Check Multiple Stops");
				thumbnailName = "";
				categoryCode = "composing-arranging-tools";
			}
		}


  MessageDialog
  {
    id: messagedialog
    visible: false
    title: "Warning"
    text: "This plugin does note take into account thumb position multiple stops on the cello.\n\nOnly staves with violin, viola or violoncello are being analysed. (Staves with sections violins, violas and violoncellos are ignored)\nSee https://musescore.org/en/node/367135\n\n\n(C)2019 Jörn Eichler (heuchi) and others\nLoosely based on the mechanism described by Ertuğrul Sevsay in his book about orchestration"
    onAccepted: {
      close();
    }
  }


 /*    
    	MessageDialog {
			id: messagedialog
			//icon: StandardIcon.Warning
      	    visible: false
		    text: "This plugin does note take into account thumb position multiple stops on the cello.\n\n\n(C)2019 Jörn Eichler (heuchi)\nLoosely based on the mechanism described by Ertuğrul Sevsay in his book about orchestration"
			standardButtons: Dialog.Ok | Dialog.Undo
			
			 Component.onCompleted: {
				var undoButton = messagedialog.addButton("Undo", DialogButtonBox.ActionRole);
				undoButton.clicked.connect(function() {
					cmd("undo");  // Handle the undo button action
					});
				}

			onAccepted: {
					(typeof(quit) === 'undefined' ? Qt.quit : quit)()
					}
			
		
		  //  onRejected: {
		  //    cmd("undo")
		  //  }
		}
 */
   /*
      MessageDialog {
    id: messagedialog
      visible: false
    text: "This plugin does not take into account thumb position multiple stops on the cello.\n\n\n(C)2019 Jörn Eichler (heuchi)\nLoosely based on the mechanism described by Ertuğrul Sevsay in his book about orchestration"
    standardButtons: Dialog.Ok | Dialog.Cancel

    onAccepted: {
        if (messagedialog.button == Dialog.Ok) {
            (typeof(quit) === 'undefined' ? Qt.quit : quit)();
        } else if (messagedialog.button == Dialog.Cancel) {
            cmd("undo");
        }
    }
}
  */  
     
      	property bool processAll: false;

      	readonly property int stopIMPOSSIBLE: 0;      
      	readonly property int stopDIFFICULT:  1;      
      	readonly property int stopPOSSIBLE:   2;      
      
        	
      	function markText(cursor, msg) {
        	var myText = newElement(Element.STAFF_TEXT);
            	myText.color = "deeppink";
            	myText.text = msg;
            	cursor.add(myText);
      	     }          
    
  
    	   function evaluate(strings, interval) {
            // find lowest and highest positions
            // open strings don't count here
            var min = 99;
            var max = 0;
            
            for(var i=0;i<strings.length;i++) {
                  if(strings[i] > 0 && strings[i] < min)
                        min = strings[i];
                  if(strings[i] > max)
                        max = strings[i];
            }
            console.log("Min = "+min+", Max = "+max);
            if(max > 0 && min < 99) {
                  // intervall needs to be no greater than
                  // a tritone on the violin
                  if(max-min > interval)
                        return stopIMPOSSIBLE;
            }
            
            // three / four equal values mean 
            // two / three equal intervals
            // which is difficult / impossible
            var allEqual = true;
            for(var i=0;i<strings.length-1 && allEqual;i++) {
                  if(strings[i] == 0
                     || strings[i] != strings[i+1])
                        allEqual = false;
            }
            if(allEqual) {
                  if(strings.length == 3)
                        return stopDIFFICULT;
                  if(strings.length == 4)
                        return stopIMPOSSIBLE;
            }

            // if a position is present more than one time
            // it needs to be on adjacent strings
            for(var i=0;i<strings.length;i++) {
                  for(var j=i+1;j<strings.length;j++) {
                        if(strings[i] != 0
                           && strings[i] == strings[j]
                           && j-i > 1)
                              return stopIMPOSSIBLE;
                  }
            }
            
            // an open string in the middle can make problems
            var state = 0;
            var hasOpenMiddle = false;
            for(var i=0;i<strings.length;i++) {
                  switch(state) {
                     case 0:
                        // look for stopped string
                        if(strings[i] != 0)
                              state=1;
                        break;
                     case 1:
                        // look for open string
                        if(strings[i] == 0)
                              state = 2;
                        break;
                     case 2:
                        // look for stopped strings
                        if(strings[i] != 0)
                              hasOpenMiddle = true;
                        break;
                  }
            }
            if(hasOpenMiddle)
                  return stopDIFFICULT;

            // if we didn't find any problems
            // the stop is possible
            return stopPOSSIBLE;
      }
      
      function findPositions(curS,notes,strings, interval, iStrings) {
          //  var instrumentStrings = [55, 62, 69, 76];
          //  var instrumenstStrings = iStrings;
            var note = notes[strings.length];
            var bestValue = stopIMPOSSIBLE;
            
            if(notes.length == strings.length) {
                  // check constraints
                  return evaluate(strings, interval);
            }
                        
            // Try string
            if(iStrings[curS] <= note.pitch) {
                  strings.push(note.pitch - iStrings[curS]);
                  console.log(strings);
                  var value = findPositions(curS+1,notes,strings, interval, iStrings);
                  strings.pop();
                  if(value == stopPOSSIBLE) return(value);
                  if(value > bestValue)
                        bestValue = value;
            }
            return(bestValue);
      }
      
      function checkStops(cursor, notes, interval, iStrings) {            
            var strings = [];
            
            // More than 4 notes -> impossible
            if(notes.length > 4) {
                  markText(cursor, "Impossible");
                  return;
            }
            
            // Try to find a possible fingering for
            // each note meeting some more constraints
            // starting at all strings
            var bestValue = stopIMPOSSIBLE;
            for(var curS=0; curS<5-notes.length; curS++) {
                  var value = findPositions(curS,notes,strings, interval, iStrings);
                  if(value > bestValue)
                        bestValue = value;
                  if(bestValue == stopPOSSIBLE)
                        break;
                  // else go on trying
            }

            switch(bestValue) {
               case stopIMPOSSIBLE:
                  markText(cursor, "IMP");
                  break;
               case stopDIFFICULT:
                  markText(cursor, "DIFF");
                  break;
               default:
            //     markText(cursor, "OK"); // debugging
                  break;
            }
      }
      	     

      	onRun: {
      	
     
      	
      	  messagedialog.open();
      	           	
				curScore.startCmd()
      		

      		
				console.log("start")
            	
          
            	            	
				// find selection
            	var startStaff;
            	var endStaff;
            	var endTick;
				var cursor = curScore.newCursor();
				cursor.rewind(1);
				
            	if (!cursor.segment) {
                  	// no selection
                  	console.log("no selection: check complete score");
                  	processAll = true;
                  	startStaff = 0;
                  	endStaff = curScore.nstaves;
					} 
				else {
                	startStaff = cursor.staffIdx;
                  	cursor.rewind(2);
                  	endStaff = cursor.staffIdx+1;
                  	endTick = cursor.tick;
                    if(endTick === 0) {
                        	// selection includes end of score
                        	// calculate tick from last score segment
                        	endTick = curScore.lastSegment.tick + 1;
                  	}
                	cursor.rewind(1);
                	console.log("Selection is: Staves("+startStaff+"-"+endStaff+") Ticks("+cursor.tick+"-"+endTick+")");
					}
            
            //	if(processAll) {
            //      msgNoSelection.visible = true;
            //    } else {
 	        	for (var staff = startStaff; staff < endStaff; staff++) {
                  		for (var voice = 0; voice < 4; voice++) {
                        		cursor.rewind(1); // sets voice to 0
                        		cursor.voice = voice; //voice has to be set after goTo
                        		cursor.staffIdx = staff;
                        		if (processAll)
                              			cursor.rewind(0) // if no selection, beginning of score
                        		while (cursor.segment && (processAll || cursor.tick < endTick)) {
                              			if (cursor.element && cursor.element.type === Element.CHORD) {
                                    		//	var graceChords = cursor.element.graceNotes;
                                    			var notes = cursor.element.notes;
                                    			if(notes.length > 1)
                                    			
   
                                    			// check instrument
                                    			var instrument;
                                    			var hasInstrumentId = curScore.staves[staff].part.instrumentId !== undefined;
												if (hasInstrumentId) {
													instrument = curScore.staves[staff].part.instrumentId
													} 
      											if (instrument === "strings.violin" || instrument === "violin") {
													var interval = 6;
													var iStrings = [55, 62, 69, 76];
													checkStops(cursor, notes, interval, iStrings);
      											} else if (instrument === "strings.viola" || instrument === "viola") {
													var interval = 5;
													var iStrings = [48, 55, 62, 65];
													checkStops(cursor, notes, interval, iStrings);
                                                } else if (instrument === "strings.cello" || instrument === "violoncello") { 
													var interval = 6;
													var iStrings = [36, 43, 50, 57];
													checkStops(cursor, notes, interval, iStrings);
      											}
      	     
                                    			
                              		}
                             		cursor.next();
                        	}
                  	} 
            	}
            	//}
           
        console.log("finished");
        curScore.endCmd()
        quit();
      	}
}
